// Rational.java 
// David Hitchcock
// Converted from  addfractions.py (5 November 2000)
// 31 October 2001
import java.util.Random;


// A class for rational numbers, or fractions
class Rational {
    private int numerator = 0;
    private int denominator = 1;
    // The static Random rand0 is used to provide a seed for rand, 
    // which should be different for each Rational.  If you don't
    // provide a seed, the time is used as a seed.  I was
    // generating two Rational's at the same time, so that they
    // got the same seed, and produced the same fraction each time
    // they generated a new problem (1/2 + 1/2 = , or 4/7 + 4/7 =,
    // for example).  This way the two numbers come out different.
    private static Random rand0 = new Random();
    private Random rand = new Random(rand0.nextLong());

    // Returns the value of the numerator
    public int numeratorof(){
	return numerator;
    }

    // Returns the value of the denominator
    public int denominatorof(){
	return denominator;
    }

// Use a random number generator to set a Rational to a
// random value within the allowed range.  If fractype is
// whole, numerator can range from 0 to max, and denominator == 1.
// If fractype is proper denominator can range from 1 to max, and
// numerator from 0 to denominator - 1.
// If fractype is improper, numerator can range from 0 to max, 
// and denominator can range from 1 to max.
// If reduced is true, the Rational is reduced to lowest 
// terms.
    public void setRandom(int max, String fractype, 
			  boolean reduced){
	if (fractype.equals("whole")) {
	    numerator = rand.nextInt(max + 1); 
		//(int)((double)(max+1) * random());
	    denominator = 1;
	    return;
	}
	else if (fractype.equals("improper")) {
	    denominator = rand.nextInt(max) + 1; 
		//(int)(((double)(max) * random())+1);
	    numerator =  rand.nextInt(max + 1); 
		//(int)((double)(max+1) * random());
	    if (reduced)
		reduce();
	    return;
	}
	else { // if (fractype.equals("proper")) or anything but
	    // "whole" and "improper"
	    denominator = rand.nextInt(max) + 1; 
		//(int)(((double)(max) * random())+1);
	    numerator = rand.nextInt(denominator);
		//(int)((double)denominator * random());
	    if (reduced)
		reduce();
	    return;
	}
    }

    public class BadStringException extends Exception {
	public String badstring;

	public BadStringException(String name) {
	    //    super("String \"" + name + "\" is not valid");
	    badstring = name;
	}
    }

    // Set a Rational to a value given by a String
    // Usage:
    // someRational.setFromString(dataString);
    // Throws BadStringException if dataString does not yield 
    // a valid rational (contains characters it shouldn't, or
    // contains no integers)
    public void setFromString(String data)
	throws BadStringException
    {
	int next = 0;
	boolean tophasdigit = false;
	boolean hasslash = false;
	boolean bottomhasdigit = false;
	// top will hold the numerator, bottom the denominator.
	StringBuffer top = new StringBuffer("");
	StringBuffer bottom = new StringBuffer("");
	int length = data.length();
	
	// Ignore blanks before the first digit	   
	while ((next < length)&&(data.charAt(next) == ' '))
	    next++;
	// Read in a '-' sign, if any, and  digits of the numerator
	if ((next < length)&&(data.charAt(next) == '-')) {
	    top.append('-');
	    next++;
	}
	// Ignore blanks before the first digit	   
	while ((next < length)&&(data.charAt(next) == ' '))
	    next++;
	while ((next < length)&&(isdigit(data.charAt(next)))) {
	    tophasdigit = true;
	    top.append(data.charAt(next));
	    next++;
	}
	// Ignore blanks before the / or end of String
	while ((next < length)&&(data.charAt(next) == ' '))
	    next++;
	// If there is a '/', we have a fraction
	if ((next < length) && (data.charAt(next) == '/')) {
	    hasslash = true;
	    next++;
	}
	// Ignore blanks before the first digit of the denominator
	while ((next < length)&&(data.charAt(next) == ' '))
	    next++;
	// Read in digits of the denominator
	while ((next < length)&&(isdigit(data.charAt(next)))) {
	    bottomhasdigit = true;
	    bottom.append(data.charAt(next));
	    next++;
	}
	// Ignore blanks after the digits of the denominator
	while ((next < length)&&(data.charAt(next) == ' '))
	    next++;
	if (next < length) 
	// String contains characters other than
	// (blanks)-(blanks)(digits)(blanks)/(blanks)(digits)(blanks)
	    throw new BadStringException("tophasdigit = " + String.valueOf(tophasdigit) + " hasslash = " + String.valueOf(hasslash) + "\nbottomhasdigit = " + String.valueOf(bottomhasdigit) + "\ntop = " + top.toString() + "  bottom = " + bottom.toString());
	if (tophasdigit && !hasslash && !bottomhasdigit) {
	// We have a whole number
	    denominator = 1;
	    numerator = Integer.parseInt(top.toString());
	    return;
	}
	if (tophasdigit && hasslash && bottomhasdigit) {
	// We have a fraction
	    int test = Integer.parseInt(bottom.toString());
	    numerator = Integer.parseInt(top.toString());
	    if (test == 0)
		throw new BadStringException("denominator == 0");
	    // because the denominator can't be 0!
	    else
		denominator = test;
	    return;
	}
	// If we reach here, input doesn't form either a valid
	// fraction or whole number, so
	throw new BadStringException("Not whole or fraction");
    }


    // Returns true if char next is a (single) digit, 
    // otherwise returns false
    public boolean isdigit(char next) {
	if (next == '0' || next == '1' || next == '2' || 
	    next == '3'	|| next == '4' || next == '5' || 
	    next == '6' || next == '7' || next == '8' || 
	    next == '9')
	    return true;
	else
	    return false;
    } 
	

    public String tostring() {
	if (denominator == 1) 
	    return (String.valueOf(numerator));
	else 
	    return (String.valueOf(numerator) + "/" + String.valueOf(denominator));
    }

    // Instruct the user on how to enter a rational number
    public String prompt() {
	return "To enter a number (addend or sum):\n" +
		"First enter the numerator (the top number),\n" +
		"followed by a '/', then the denominator\n" +
		"(the bottom number).\n" +
		"To enter a whole number, type the number,\n" +
		"then hit 'enter.'\n" +
		"To subtract, put a \"-\" in front of the\n" +
		"second number.";
    }

// Reduce a Rational to lowest terms 
    public void reduce() {
	if (numerator == 0) {
	    denominator = 1;
	    return;
	}
	int max = numerator;
	if (denominator < numerator)
	    max = denominator;
	int factor = 2;
        while (factor <= max) {
	    while ((numerator % factor == 0) &&
		       (denominator % factor == 0)){
			   numerator = numerator / factor;
			   denominator = denominator / factor;
			   max = max / factor;
	    }
	    factor++;
	}
    }

    // Returns the sum of two Rationals
    // Usage: sum = rat1.plus(rat2);
    public Rational plus(Rational rat2) { 
	Rational sum = new Rational();
	sum.numerator = (numerator * rat2.denominator +  
		     rat2.numerator * denominator);
	sum.denominator = denominator * rat2.denominator;
	sum.reduce();
	return sum;
    }

    // Returns the difference of two Rationals
    // Usage: difference = rat1.minus(rat2);
    public Rational minus(Rational rat2) {
	Rational difference = new Rational();
	difference.numerator = (numerator * rat2.denominator - 
			     rat2.numerator * denominator);
	difference.denominator = denominator * rat2.denominator;
	difference.reduce();
	return difference;
    }

    // Returns the product of two Rationals
    // Usage: produce = rat1.times(rat2);
    public Rational times(Rational rat2) {
	Rational product = new Rational();
	product.numerator = numerator * rat2.numerator;
	product.denominator = denominator * rat2.denominator;
	product.reduce();
	return product;
    }

    // Returns the quotient of rat1 divided by rat2.
    // Usage: quotient = rat1.dividedby(rat2);
    public Rational dividedby(Rational rat2) {
	Rational quotient = new Rational(); 
	quotient.numerator = numerator * rat2.denominator;
	quotient.denominator = denominator * rat2.numerator;
	quotient.reduce();
	return quotient;
    }

    // Compare two Rationals.
    // usage: if (rat.isEqual(rat1)) {
    public boolean isEqual(Rational rat1) {
	if ((numerator == rat1.numerator) && 
			(rat1.denominator == denominator))
	    return true;
	else
	    return false;
    }

    // Compare two Rationals.
    // Returns 'true' even if one or both are not in lowest terms
    // usage: if (rat.isEquivalent(rat1)) {
    public boolean isEquivalent(Rational rat1) {
	if ((numerator * rat1.denominator) == 
			(rat1.numerator * denominator))
	    return true;
	else
	    return false;
    }

    // Add both the numerator and denominator		
    // (This is quite a common error)
    // Usage: wronganswer = rat1.addboth(rat2);
    public Rational addboth(Rational rat2) {
	Rational wronganswer = new Rational();
	wronganswer.numerator = numerator + rat2.numerator;
	wronganswer.denominator = denominator + rat2.denominator;
	wronganswer.reduce();
	return wronganswer;
    }

    // Subtract both the numerator and denominator		
    // (This is also not unusual)
    // Usage: wronganswer = rat1.addboth(rat2);
    public Rational subtractboth(Rational rat2) {
	Rational wronganswer = new Rational();
	wronganswer.numerator = numerator - rat2.numerator;
	wronganswer.denominator = denominator - rat2.denominator;
	wronganswer.reduce();
	return wronganswer;
    }
}
